/*
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "pim/mail/MailSyncSource.h"
#include "spds/SyncItemStatus.h"
#include "spds/MailMessage.h"
#include "spds/constants.h"
#include "base/util/utils.h"
#include "base/util/StringBuffer.h"
#include "pim/utils.h"

static int call;
static int cnew;
static int cupdated;
static int cdeleted;


MailSyncSource::MailSyncSource(const wchar_t* name) : SyncSource(name)
{    
    path = stringdup(TEXT("\\"));
    s_folderToSync = NULL;   
    syncItem = NULL;
    isMailInOutbox = false;
    isMailInInbox = false;
    mailClient2Server = NULL;
    mailServer2Client = NULL;
}

MailSyncSource::MailSyncSource(MailSyncSource& s) : SyncSource(s)
{
    setLastSync(s.getLastSync());
    setNextSync(s.getNextSync());
    setRemoteURI(s.getRemoteURI());
    setType(s.getType());
    setLastAnchor(s.getLastAnchor());
    setNextAnchor(s.getNextAnchor());
    setErrorHandler(s.getErrorHandler());
    setFilter(s.getFilter());

    setPath(s.path);
    setFolderToSync(s.s_folderToSync);

    // These are not copied from the other instance
    isMailInOutbox = false;
    isMailInInbox = false;
    mailServer2Client=NULL;
    mailClient2Server=NULL;
    extras.clear();
}

MailSyncSource::~MailSyncSource() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;
    }
    if (s_folderToSync) {
        delete [] s_folderToSync; s_folderToSync = NULL;
    }
    if (path) {
        delete [] path; path = NULL;
    }
    if (mailServer2Client)
        delete mailServer2Client;
    if (mailClient2Server)
        delete mailClient2Server;
}

void MailSyncSource::setPath(const wchar_t* p) {
    if (path)
        delete [] path;

    if (p == NULL)
        path = stringdup(TEXT(""));
    else
        path = stringdup(p);
}

const wchar_t* MailSyncSource::getPath() {
    return path;
}

void MailSyncSource::setFolderToSync(const wchar_t* t){
    safeDelete(&s_folderToSync);
    s_folderToSync = stringdup(t);
}

const wchar_t* MailSyncSource::getFolderToSync() {
    return s_folderToSync;
}

bool MailSyncSource::isInFolderToSync(const wchar_t* toCheck) {
    return wcsstr(s_folderToSync, toCheck) != NULL ? true : false;
}

void MailSyncSource::setIsMailInOutbox(bool value){
    isMailInOutbox = value;
    
}

bool MailSyncSource::getIsMailInOutbox() {
    return isMailInOutbox;
}

void MailSyncSource::setIsMailInInbox(bool value){
    isMailInInbox = value;    
}

bool MailSyncSource::getIsMailInInbox() {
    return isMailInInbox;
}

/*
 * Return the first SyncItem of all. It is used in case of slow 
 * sync and retrieve the entire data source content.
 */
SyncItem* MailSyncSource::getFirstItem()
{
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    //
    // initialize the flag
    //
    isMailInInbox = false;

    if ( mailClient2Server->setSync(SYNC_SLOW, getPath(), extras) != S_OK ){
        logError(TEXT("getFirstItem: error in MailClient2Server::setSync"));
        return NULL;
    }
    
    MailClientData* m = mailClient2Server->getFirstItem();
    
    if (m) {              
        
        MailMessage& mailItem = m->getEmailData()->getEmailItem();
        //
        // get the MailExtraData object to retrieve i.e. message-id... It is only the 
        // pointer. don't destroy it!
        //
        const MailExtraData* med = getMailExtraData(extras, mailItem.getEntryID());
        
        // extra is empty because the previous file doesn't exist or is empty,
        // or is a slow after previous fast
        if (med) {
            mailItem.setMessageId(med->getMessageId());
        }
        else {
            MailExtraData m;
            wchar_t *id = createMessageID(mailItem.getEntryID());

            m.setId(mailItem.getEntryID());
            m.setMessageId(id);
            extras.add(m);

            mailItem.setMessageId(id);

            delete [] id;
        }
        
        syncItem = new SyncItem(mailItem.getEntryID());        

        wchar_t* s = m->getEmailData()->format();        
        char* toSend = toMultibyte(s);
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));                    
        syncItem->setState(SYNC_STATE_NONE);

        syncItem->setSourceParent(m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        if (s) {delete [] s; s = NULL; }
        if (toSend) {delete [] toSend; toSend = NULL; }        
        logInfo(TEXT("Subject mail get: -%s-"), m->getEmailData()->getEmailItem().getSubject());
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No mail get"));
    }
    
    return syncItem;    
}

/*
* Return the next SyncItem of all. It is used in case of slow or refresh sync 
* and retrieve the entire data source content.
*/

SyncItem* MailSyncSource::getNextItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextItem();   
    if (m) {      
        
        MailMessage& mailItem = m->getEmailData()->getEmailItem();      
        //
        // get the MailExtraData object to retrieve i.e. message-id...
        // It must exist after getFirstItem

        const MailExtraData* med = getMailExtraData(extras, mailItem.getEntryID());                
                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        }
        else {
            MailExtraData m;
            wchar_t *id = createMessageID(mailItem.getEntryID());

            m.setId(mailItem.getEntryID());
            m.setMessageId(id);
            extras.add(m);

            mailItem.setMessageId(id);

            delete [] id;
        }
        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());
        wchar_t* s = m->getEmailData()->format();
        logDebug(s);
        char* toSend = toMultibyte(s);
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));            
        syncItem->setState(SYNC_STATE_NONE);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
                
        if (s) {delete [] s; s = NULL; }
        if (toSend) {delete [] toSend; toSend = NULL; }
        
        logInfo(TEXT("Subject mail get: -%s-"), m->getEmailData()->getEmailItem().getSubject());
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No mail get"));
    }

    return syncItem;  

}

SyncItem* MailSyncSource::getFirstItemKey() {
    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    } 
    //
    // initialize the flag
    //
    setIsMailInOutbox(false);
    
    mailClient2Server->setFolderToSync(s_folderToSync);

    if ( mailClient2Server->setSync(SYNC_SLOW, getPath(), extras) != S_OK ){
        logError(TEXT("getFirstItemKey: error in MailClient2Server::setSync"));
        return NULL;
    }

    MailClientData* m = mailClient2Server->getFirstItem();
    
    if (m) {      
        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());
        delete m; m = NULL;
    }
    
    return syncItem;    
    
}

SyncItem* MailSyncSource::getNextItemKey() {
    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextItem();   
    if (m) {      
        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());
        delete m; m = NULL;
    }

    return syncItem;  
}

/*
* This method is always called by the sync manager. So the setModifiedItemsPOOM is always called
*/

SyncItem* MailSyncSource::getFirstNewItem() {

    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    //
    // initialize the flag
    //
    setIsMailInOutbox(false);
    
    mailClient2Server->setFolderToSync(s_folderToSync);

    if ( mailClient2Server->setSync(SYNC_TWO_WAY, getPath(), extras) != S_OK ){
        logError(TEXT("getFirstNewItem: error in MailClient2Server::setSync"));
        return NULL;
    }
    
    MailClientData* m = mailClient2Server->getFirstNewItem();   
    
    if (m) {      

        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        //
        // get the MailExtraData object to retrieve i.e. message-id...
        //
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());
        
        if (med) {
            emailItem.setMessageId(med->getMessageId());
        }
        else {
                    
            MailExtraData m;
            wchar_t *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }
        
        syncItem = new SyncItem(emailItem.getEntryID());
        
        wchar_t* s = m->getEmailData()->format();
        logDebug(s);
        char* toSend = toMultibyte(s);
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));      
        syncItem->setState(SYNC_STATE_NEW);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        

        if (s) {delete [] s; s = NULL; }
        if (toSend) {delete [] toSend; toSend = NULL; }
        
        logInfo(TEXT("Subject first new mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());
        delete m; m = NULL;

    } else {
        logInfo(TEXT("No new mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getNextNewItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }
    
    MailClientData* m = mailClient2Server->getNextNewItem();   
    
    if (m) {      
        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());                                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        } 
        else {
                    
            MailExtraData m;
            wchar_t *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }
        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());
        
        wchar_t* s = m->getEmailData()->format();
        logDebug(s);
        char* toSend = toMultibyte(s);
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));              
        syncItem->setState(SYNC_STATE_NEW);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        if (m->getParent()[0] == 'O')
            isMailInOutbox = true;
        
        if (s)
            {delete [] s; s = NULL; }
        if (toSend) {delete [] toSend; toSend = NULL; }
        
        logInfo(TEXT("Subject other new mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());
        delete m; m = NULL;
        
    } else {
        logInfo(TEXT("No other new mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getFirstUpdatedItem() {
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
        
    MailClientData* m = mailClient2Server->getFirstUpdatedItem();   
    
    if (m) {      
        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());                                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        } 
        else {
                    
            MailExtraData m;
            wchar_t *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }
        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());
        
        wchar_t* s = m->getEmailData()->format();
        logDebug(s);
        char* toSend = toMultibyte(s);
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));       
        syncItem->setState(SYNC_STATE_UPDATED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        if (s) {delete [] s; s = NULL; }
        if (toSend) {delete [] toSend; toSend = NULL; }
        
        logInfo(TEXT("Subject first updated mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());
        delete m; m = NULL;
        
    } else {
        logInfo(TEXT("No updated mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getNextUpdatedItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getNextUpdatedItem();   
    
    if (m) {      
        
        MailMessage& emailItem = m->getEmailData()->getEmailItem();
        const MailExtraData* med = getMailExtraData(extras, emailItem.getEntryID());                                
        if (med) {
            m->getEmailData()->getEmailItem().setMessageId(med->getMessageId());               
        } 
        else {
                    
            MailExtraData m;
            wchar_t *msgId = createMessageID(emailItem.getEntryID());

            m.setId(emailItem.getEntryID());
            m.setMessageId(msgId);
            extras.add(m);

            emailItem.setMessageId(msgId);
            delete [] msgId;
        }

        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());
        
        wchar_t* s = m->getEmailData()->format();
        logDebug(s);
        char* toSend = toMultibyte(s);
        syncItem->setData(toSend, strlen(toSend) * sizeof(char));       
        syncItem->setState(SYNC_STATE_UPDATED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        if (s) {delete [] s; s = NULL; }
        if (toSend) {delete [] toSend; toSend = NULL; }
        
        logInfo(TEXT("Subject other updated mail: -%s-"), m->getEmailData()->getEmailItem().getSubject());
        delete m; m = NULL;
        
    } else {
        logInfo(TEXT("No other updated mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getFirstDeletedItem() {
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getFirstDeletedItem();   
    
    if (m) {      
        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());      
        syncItem->setState(SYNC_STATE_DELETED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        
        logInfo(TEXT("ID of deleted mail: -%s-"), m->getEmailData()->getEmailItem().getEntryID());
        delete m; m = NULL;
    
    } else {
        logInfo(TEXT("No deleted mail"));
    }

    return syncItem;  
}

SyncItem* MailSyncSource::getNextDeletedItem() {    
    if (syncItem) {
        delete syncItem; syncItem = NULL;   
    }  
    
    MailClientData* m = mailClient2Server->getNextDeletedItem();   
    
    if (m) {      
        syncItem = new SyncItem(m->getEmailData()->getEmailItem().getEntryID());     
        syncItem->setState(SYNC_STATE_DELETED);
        syncItem->setSourceParent((wchar_t*)m->getParent());
        logInfo(TEXT("ID of other deleted mail: -%s-"), m->getEmailData()->getEmailItem().getEntryID());

        delete m; m = NULL;

    } else {
        logInfo(TEXT("No other deleted mail"));
    }

    return syncItem;  
}

void MailSyncSource::setItemStatus(const wchar_t* key, int status) {
    logDebug(TEXT("key: %s, status: %i"), key, status);
}

int MailSyncSource::addItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    MailClientData m;
    MailExtraData med;
    wchar_t* id = NULL;
    wchar_t tt[3];
    wchar_t *d = NULL;
    int errorCode = 0;
    char* tmp = NULL;
    wchar_t* toUse = NULL;
    
    mailServer2Client->setFolderToSync(s_folderToSync);
    
    if (item.getTargetParent() == NULL) {
        wsprintf(tt, TEXT("%c"), item.getKey()[0]);
    }  else {
        wsprintf(tt, TEXT("%s"), item.getTargetParent());
    }
    
    if (!isInFolderToSync(tt))
        goto finally;
    
    tmp = new char[item.getDataSize() + 1];
    memset(tmp, 0, item.getDataSize() + 1);
    memcpy(tmp, item.getData(), item.getDataSize());

    toUse = toWideChar(tmp);
    logDebug(toUse);
    delete [] tmp; 
    tmp = NULL;

    errorCode = m.emailData->parse(toUse);

    delete [] toUse; 
    toUse = NULL;

    if (errorCode != 0) {
        logError(TEXT("Error parsing message!"));
        ret= STC_COMMAND_FAILED; // command Failed;
        goto finally;
    }    
    m.setParent(tt);
    
    id = mailServer2Client->addMessage(&m);
    
    if (id == NULL) {
        logError( TEXT("Error in MailServer2Client::addMessage") );
        goto finally;
    }
    
    // Ok: set the new key and the status code
    ret = STC_ITEM_ADDED;        
    item.setKey(id);
    logInfo(TEXT("Mail added: -%-"), m.getEmailData()->getEmailItem().getSubject());
    
    // set all extra data
    //
    med.setId(id);
    med.setMessageId(m.emailData->getEmailItem().getMessageId());
    extras.add(med);

    // Set new mail in Inbox flag
    if (isMailInInbox == false && wcscmp(tt, TEXT("I")) == 0) 
        isMailInInbox = true;

finally:
    
    if (d) {
        delete [] d; d = NULL;
    }
    return ret;
}

int MailSyncSource::updateItem(SyncItem& item) {
    
    int ret = STC_COMMAND_FAILED; // command Failed;
    wchar_t tt[3];
    MailClientData m; 
    wchar_t *d = NULL;
    int errorCode = 0;
    HRESULT hr = S_OK;
    char* tmp = NULL;
    wchar_t* toUse = NULL;

    mailServer2Client->setFolderToSync(s_folderToSync);
          
     if (!isInFolderToSync(tt))
        goto finally;     
    
    tmp = new char[item.getDataSize() + 1];
    memset(tmp, 0, item.getDataSize() + 1);
    memcpy(tmp, item.getData(), item.getDataSize());

    toUse = toWideChar(tmp);

    delete [] tmp; 
    tmp = NULL;

    errorCode = m.emailData->parse(toUse);
    
    delete [] toUse;
    toUse = NULL;

    if (errorCode != 0) {
        // failure
        goto finally;
    }
    
    m.setParent(tt);
           
    hr = mailServer2Client->updateMessage(&m);
    
    if (hr == S_OK) {        
        ret = STC_OK;     
        logInfo(TEXT("Mail updated: -%-"), m.getEmailData()->getEmailItem().getSubject());
    }    
    
finally:
    
    if (d) {
        delete [] d; d = NULL;
    }
    return ret;
}

int MailSyncSource::deleteItem(SyncItem& item) {
    
   int ret = STC_COMMAND_FAILED; // command Failed;
    
    wchar_t* id = NULL;
    wchar_t tt[3];

    mailServer2Client->setFolderToSync(s_folderToSync);
    MailClientData m;

    if (item.getTargetParent() == NULL) {
        wsprintf(tt, TEXT("%c"), item.getKey()[0]);
    }  else {
        wsprintf(tt, TEXT("%s"), item.getTargetParent());    
    }

    m.setParent(tt);

    m.emailData->getEmailItem().setEntryID(item.getKey());
    
    HRESULT hr = mailServer2Client->deleteMessage(&m);
    
    if (hr == S_OK) {        
        ret = STC_OK;          
        logInfo(TEXT("Mail deleted: -%-"), item.getKey());
    }
        
    return ret;
}

int MailSyncSource::beginSync() {
    // Init mailServer2Client
    if (mailServer2Client)
        delete mailServer2Client;

    mailServer2Client = new MailServer2Client();
    
    if(!mailServer2Client)
        return 500;

    // Init mailClient2server
    if (mailClient2Server)
        delete mailClient2Server;

    mailClient2Server = new MailClient2Server();
    mailClient2Server->setFolderToSync(s_folderToSync);
        
    if(!mailClient2Server)
        return 500;

    return 0;
}

int MailSyncSource::endSync() {
    
    // Get a new mailClient2server
    if (mailClient2Server)
        delete mailClient2Server;

    mailClient2Server = new MailClient2Server();
    mailClient2Server->setFolderToSync(s_folderToSync);
    mailClient2Server->setSync(ONLY_ALL_MESSAGE_LIST, getPath(), extras);    
    mailClient2Server->writeCurrentMails(getPath(), extras);

    return 0;
}

MailSyncSource& MailSyncSource::operator=(MailSyncSource& s) {
    if (this != &s) {
        setSyncMode(s.getSyncMode());
        setLastSync(s.getLastSync());
        setNextSync(s.getNextSync());
        setRemoteURI(s.getRemoteURI());
        setType(s.getType());
        setLastAnchor(s.getLastAnchor());
        setNextAnchor(s.getNextAnchor());
        setErrorHandler(s.getErrorHandler());

        setFilter(s.getFilter());
        setPath(s.getPath());
        setFolderToSync(s.getFolderToSync());

        // This attributes are not copied from the other source
        setIsMailInOutbox(false);
        setIsMailInInbox(false);
        mailClient2Server = NULL;
        mailServer2Client = NULL;
        extras.clear();
    }
    return *this;
}

ArrayElement* MailSyncSource::clone() {
    MailSyncSource* s = new MailSyncSource(getName());
    *s = *this;
    return s;
}



